/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* USER DATABASE UPDATE ROUTINES
 *
 * This contains code to modify the password and full-name fields in the user
 * database.  Nothing is included here to modify uid, gid, or directory fields.
 * Some of the code here would help support that, but such functions would
 * never be built into the interpretor, so they won't go in this file, which
 * is meant to be shared between the interpretor and the administration tools.
 */

#include "backtalk.h"

#include <pwd.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#include "lock.h"
#include "str.h"
#include "udb_check.h"
#include "udb_edit.h"
#include "udb_read.h"
#include "set_uid.h"

#if defined(AUTH_DBM) || defined(UD_DBM) || defined(UID_MAP_FILE)
#include "hashfile.h"
#endif

void killdir();

#if defined(AUTH_FILE) || defined(UD_PASSWD)

/* UPDATE_PW - This updates flat passwd-type files.  It opens the given file,
 * and searches it for a line starting with the given key and a colon.  If
 * replace is not NULL, it replaces that line with the one given, otherwise
 * it deletes that line.  We check for full disks and such problems.  It is
 * the responsibility of the caller to set our uid to something that has write
 * access to the file and the directory containing the file.  The replace line
 * should include a trailing newline.  Returns 1 if there was no such user.
 */

int update_pw(char *file, char *key, char *replace)
{
char tmp[BFSZ], bf[BFSZ+1];
FILE *tmpfp, *fp;
struct stat st;
int tmpfd, keylen;
int did_change= 0;
long size;
int oldumask;

    /* Open original password file */

    if ((fp= fopen(file,"r")) == NULL)
	die ("Cannot open %s",file);
    lock_exclusive(fileno(fp),file);
    fstat(fileno(fp), &st);
    size= st.st_size;

    /* Open temporary file */

    sprintf(tmp,"%s.tmp",file);
    oldumask= umask(0);
    tmpfd= open(tmp, O_WRONLY|O_CREAT|O_EXCL, st.st_mode);
    umask(oldumask);
    if (tmpfd < 0)
    {
	fclose(fp);
	unlock(file);
	if (errno == EEXIST)
	    die("%s already exists -- I won't overwrite it",tmp);
	else
	    die("cannot create %s",tmp);
    }
    tmpfp= fdopen(tmpfd, "w");

    /* Copy the original over into the temporary, looking for update line */

    keylen= strlen(key);
    while (fgets(bf,BFSZ,fp) != NULL)
    {
	if (bf[keylen] == ':' && !strncmp(bf,key,keylen))
	{
	    did_change++;
	    size-= strlen(bf);
	    if (replace != NULL)
	    {
		fputs(replace,tmpfp);
		size+= strlen(replace);
	    }
	}
	else
	    fputs(bf,tmpfp);
    }

    if (!did_change)
    {
	unlink(tmp);
	fclose(fp);
	unlock(file);
	return 1;
    }

    /* Check to make sure that the writes all succeeded */
    fflush(tmpfp);
    fstat(fileno(tmpfp), &st);
    if (st.st_size != size)
    {
	unlink(tmp);
	fclose(fp);
	unlock(file);
	die("Error updating %s.  Disk full?",file);
    }

    /* Replace the old file with the temp file */
    unlink(file);
    if (!link(tmp,file))
	unlink(tmp);

    fclose(fp);
    unlock(file);
    return 0;
}
#endif /* AUTH_FILE || UD_PASSWD */


/* SAVE_TAG - Write the given variable/value pair into a tag file, replacing
 * the old value, if there was one, or appending the value.  The value is
 * not changed if it was protected, in which case a non-zero value is returned.
 * This always takes four arguments, but the last may be either an integer or
 * a string depending on the type.  This is not in #ifdef UD_TAGFILE, becuase
 * it might be useful for other tagfiles.
 */

#if __STDC__ && HAVE_STDVARARGS
int save_tag(char *fname, int amadm, char *varname, int type, ...)
#else
int save_tag(fname, amadm, varname, type, va_alist)
    char *fname;
    int amadm;
    char *varname;
    int type;
    va_dcl
#endif
{
#if HAVE_STDVARARGS
va_list ap;
#endif
char *bf, *val, *end, *var;
FILE *fp;
int i, protect, done= 0, rc= 0;

    bf= read_file(fname);
    VA_START(ap, type);

    /* Truncate and prepare to rewrite */
    if ((fp= fopen(fname,"w+")) == NULL)
	die("Cannot open file %s to write",fname);
    lock_exclusive(fileno(fp),fname);

    /* Process each line of saved file contents, writing back or replacing */
    for (var= bf; var != NULL && *var != '\0'; var= end+1)
    {
	/* Parse file line */
	if (protect= (*var == '*')) var++;
	val= strchr(var,'=');
	end= strchr(var,'\n');
	if (val != NULL && end != NULL && val > var && end > val)
	{
	    *(val++)= '\0';
	    *end= '\0';
	    if (!done && !strcmp(var,varname))
	    {
		done= 1;
		if (protect && !amadm)
		    rc= 1;
		else
		{
		    if (type == TK_INTEGER)
			fprintf(fp,"%s%s=%d\n", protect?"*":"",
			    var, VA_ARG1(ap, int));
		    else
			fprintf(fp,"%s%s={%s}\n", protect?"*":"",
			    var, VA_ARG1(ap, char *));
		    continue;
		}
	     }
	     fprintf(fp,"%s=%s\n", var-protect, val);
	}
    }

    /* Process the variable if not previously in file */
    if (!done)
    {
	if (type == TK_INTEGER)
	    fprintf(fp,"%s=%d\n", var, VA_ARG1(ap, int));
	else
	    fprintf(fp,"%s={%s}\n", var, VA_ARG1(ap, char *));
    }

    fclose(fp);
    unlock(fname);
    if (bf != NULL) free(bf);
    VA_END(ap);
    return rc;
}


/* CHANGE_PASSWD - change the user's password to the given value.  Pass in
 * the encrypted password and the new status.
 */

#ifdef AUTH_FILE
void change_passwd(struct passwd *pw, char *encpw)
{
char bf[BFSZ];

    sprintf(bf,"%s:%s\n",pw->pw_name, encpw);

    set_httpd();
    update_pw(AUTH_FILE, pw->pw_name, bf);
    set_cfadm();

}
#endif /* AUTH_FILE */

#if defined(SHARE_UPW) && defined(UD_PASSWD) && defined(PASSWD_FILE)
void change_passwd(struct passwd *pw, char *encpw)
{
char bf[BFSZ];

    sprintf(bf,"%s:%s:%d:%d:%s:%s:\n",
	    pw->pw_name, encpw,
	    pw->pw_uid, pw->pw_gid,
	    pw->pw_gecos, pw->pw_dir);

    set_httpd();
    update_pw(PASSWD_FILE, pw->pw_name, bf);
    set_cfadm();
}
#endif /* SHARE_UPW && UD_PASSWD && PASSWD_FILE */

#ifdef AUTH_DBM
void change_passwd(struct passwd *pw, char *encpw)
{
int rc;

    set_httpd();

    rc= storedbm(AUTH_DBM,
    		 pw->pw_name, strlen(pw->pw_name),
    		 encpw, strlen(encpw)+1,
    		 0);
    set_cfadm();

    if (rc) die("cannot open user database "AUTH_DBM);
}
#endif /* AUTH_DBM */

#if defined(SHARE_UPW) && defined(UD_DBM)
void change_passwd(struct passwd *pw, char *encpw)
{
char bf[BFSZ];
int rc;

    set_httpd();

    sprintf(bf,"%s:%d:%d:%s:%s:\n",
	    encpw, pw->pw_uid, pw->pw_gid,
	    pw->pw_gecos, pw->pw_dir);

    rc= storedbm(DBM_NAM_FILE,
    		 pw->pw_name, strlen(pw->pw_name),
    		 bf, strlen(bf)+1,
    		 0);
    set_cfadm();

    if (rc) die("cannot open user database "DBM_NAM_FILE);
}
#endif /* SHARE_UPW && UD_DBM */


/* CHANGE_FNAME - change the user's fullname to the given value.
 */

#if defined(UD_PASSWD) && defined(PASSWD_FILE)
void change_fname(struct passwd *pw, char *newfname)
{
char bf[BFSZ];

#ifdef SHARE_UPW
    sprintf(bf,"%s:%s:%d:%d:%s:%s:\n", pw->pw_name, pw->pw_passwd,
	pw->pw_uid, pw->pw_gid, newfname, pw->pw_dir);
#else
    sprintf(bf,"%s:*:%d:%d:%s:%s:\n", pw->pw_name,
	pw->pw_uid, pw->pw_gid, newfname, pw->pw_dir);
#endif

    set_httpd_if_shared();
    update_pw(PASSWD_FILE, pw->pw_name, bf);
    set_cfadm_if_shared();

    flushusercache();
}
#endif /* UD_PASSWD && PASSWD_FILE */

#ifdef UD_DBM
void change_fname(struct passwd *pw, char *newfname)
{
char bf[BFSZ];
int rc;

#ifdef SHARE_UPW
    sprintf(bf,"%s:%d:%d:%s:%s:\n", pw->pw_passwd,
	pw->pw_uid, pw->pw_gid, newfname, pw->pw_dir);
#else
    sprintf(bf,"*:%d:%d:%s:%s:\n",
	pw->pw_uid, pw->pw_gid, newfname, pw->pw_dir);
#endif

    set_httpd_if_shared();
    rc= storedbm(DBM_NAM_FILE,
    		 pw->pw_name, strlen(pw->pw_name),
    		 bf, strlen(bf)+1, 0);
    set_cfadm_if_shared();

    flushusercache();

    if (rc) die("Cannot open user database "DBM_NAM_FILE" (%s)",errdbm());
}

#endif /* UD_DBM */

#ifdef UD_TAGFILE

void change_fname(struct passwd *pw, char *newfname)
{
char *fname;

    /* Construct name of ident file */
    fname= gethomedir(pw->pw_name);
    strcat(fname,"/");
    strcat(fname,IDENT_FILE);

    flushusercache();

    if (save_tag(fname, 0, "fullname", TK_STRING, newfname))
	    die("Protected fullname field in %s",fname);
}

#endif /* UD_TAGFILE */


/* CHANGE_GID - change the user's groupid to the given value.
 */

#if defined(UD_PASSWD) && defined(PASSWD_FILE)
void change_gid(struct passwd *pw, gid_t newgid)
{
char bf[BFSZ];

#ifdef SHARE_UPW
    sprintf(bf,"%s:%s:%d:%d:%s:%s:\n", pw->pw_name, pw->pw_passwd,
	pw->pw_uid, newgid, pw->pw_gecos, pw->pw_dir);
#else
    sprintf(bf,"%s:*:%d:%d:%s:%s:\n", pw->pw_name,
	pw->pw_uid, newgid, pw->pw_gecos, pw->pw_dir);
#endif

    set_httpd_if_shared();
    update_pw(PASSWD_FILE, pw->pw_name, bf);
    set_cfadm_if_shared();

    flushusercache();
}
#endif /* UD_PASSWD && PASSWD_FILE */

#ifdef UD_DBM
void change_gid(struct passwd *pw, gid_t newgid)
{
char bf[BFSZ];
int rc;

#ifdef SHARE_UPW
    sprintf(bf,"%s:%d:%d:%s:%s:\n", pw->pw_passwd,
	pw->pw_uid, newgid, pw->pw_gecos, pw->pw_dir);
#else
    sprintf(bf,"*:%d:%d:%s:%s:\n",
	pw->pw_uid, newgid, pw->pw_gecos, pw->pw_dir);
#endif

    set_httpd_if_shared();
    rc= storedbm(DBM_NAM_FILE,
    		 pw->pw_name, strlen(pw->pw_name),
    		 bf, strlen(bf)+1, 0);
    set_cfadm_if_shared();

    flushusercache();

    if (rc) die("Cannot open user database "DBM_NAM_FILE" (%s)",errdbm());
}

#endif /* UD_DBM */

#ifdef UD_TAGFILE

void change_gid(struct passwd *pw, gid_t newgid)
{
char *fname;

    /* Construct name of ident file */
    fname= gethomedir(pw->pw_name);
    strcat(fname,"/");
    strcat(fname,IDENT_FILE);

    flushusercache();

    save_tag(fname, 1, "gid", TK_INTEGER, newgid);
}

#endif /* UD_TAGFILE */


/* KILL_USER - This deletes the user from whatever databases he is in and
 * removes his home directory and all his files.  Returns nonzero if there
 * was no such user, 0 on success.
 */

int kill_user(struct passwd *pw)
{
int namelen= strlen(pw->pw_name);
int rc;

#ifdef AUTH_FILE
    /* Delete an entry from a flat .htpasswd file */
    set_httpd();
    rc= update_pw(AUTH_FILE, pw->pw_name, NULL);
    set_cfadm();
#endif /* AUTH_FILE */

#ifdef AUTH_DBM
    /* Delete an entry from an apache .htpasswd dbm file */
    set_httpd();
    rc= deletedbm(AUTH_DBM, pw->pw_name, namelen, 0);
    set_cfadm();
#endif /* AUTH_DBM */

#ifndef SHARE_UPW
#ifdef PASSWD_FILE
    /* Delete an entry from a flat passwd file */
    update_pw(PASSWD_FILE, pw->pw_name, NULL);
#endif /* PASSWD_FILE */

#ifdef DBM_NAM_FILE
    /* Delete an entry from a by-name dbm password file */
    deletedbm(DBM_NAM_FILE, pw->pw_name, namelen, 0);
#endif /* DBM_NAM_FILE */
#endif /* SHARE_UPW */

#ifdef DBM_UID_FILE
    /* Delete an entry from a by-uid dbm password file */
    deletedbm(DBM_UID_FILE, (char *)&pw->pw_uid, sizeof(pw->pw_uid), 0);
#endif /* DBM_UID_FILE */

#ifdef UID_MAP_FILE
    /* Delete an entry from the mapuid file
     * Note that this file is indexed by the user's realuid, which is
     * different from the one in pw->pw_uid, so we can't use that to
     * find the key to delete.  So we have to do a search through the
     * whole database for one whose content matches our name key.
     */
    deletedbm(UID_MAP_FILE, pw->pw_name, namelen+1, 1);
#endif /* UID_MAP_FILE */

    killdir(pw->pw_dir);

    return rc;
}


/* KILLDIR - Remove a directory and all files below it. */

void killdir(char *dir)
{
pid_t cpid, wpid;

    if ((cpid= fork()) == 0)
    {
	execl("/bin/rm","rm","-rf",dir,(char *)NULL);
	exit(1);
    }

    while ((wpid= wait((int *)0)) != cpid && wpid != -1)
	;
}
